@@ -240,18 +240,27 @@ impl<'a> VariantConfig<'a> {
     pub fn decode(&self, name: &syn::Ident, context: usize) -> proc_macro2::TokenStream {
         let tag = self.tag(context);
         let ident = &self.variant.ident;
+        let is_explicit = self.tag.as_ref().map_or(false, |tag| tag.explicit);
+        let explicit_wrapper = |tag, decode| quote!(decoder.decode_sequence(#tag, #decode));
 
-        let encode = match &self.variant.fields {
+        match &self.variant.fields {
             syn::Fields::Unit => {
-                quote!(if let Ok(()) = <()>::decode_with_tag(decoder, #tag) { return Ok(#name::#ident) })
+                let decode_op = if is_explicit {
+                    (explicit_wrapper)(tag, quote!(<()>::decode))
+                } else {
+                    quote!(<()>::decode_with_tag(decoder, #tag))
+                };
+
+                quote!(if let Ok(()) = #decode_op { return Ok(#name::#ident) })
             }
             syn::Fields::Unnamed(_) => {
                 if self.variant.fields.len() != 1 {
                     panic!("Tuple struct variants should contain only a single element.");
                 }
 
-                let decode_operation = if self.container_config.automatic_tags || self.tag.is_some()
-                {
+                let decode_operation = if is_explicit {
+                    (explicit_wrapper)(tag, quote!(|decoder| <_>::decode(decoder)))
+                } else if self.container_config.automatic_tags || self.tag.is_some() {
                     quote!(<_>::decode_with_tag(decoder, #tag))
                 } else {
                     quote!(<_>::decode(decoder))
@@ -267,6 +276,12 @@ impl<'a> VariantConfig<'a> {
                     field_config.decode(name, i)
                 });
 
+                let decode_op = if is_explicit {
+                    (explicit_wrapper)(tag.clone(), quote!(decode_fn))
+                } else {
+                    quote!((decode_fn)(decoder))
+                };
+
                 quote! {
                     let decode_fn = |decoder: &mut D| {
                         decoder.decode_sequence(#tag, |decoder| {
@@ -274,19 +289,9 @@ impl<'a> VariantConfig<'a> {
                         })
                     };
 
-                    if let Ok(value) = (decode_fn)(decoder) { return Ok(value) }
+                    if let Ok(value) = #decode_op { return Ok(value) }
                 }
             }
-        };
-
-        if self.tag.as_ref().map_or(false, |tag| tag.explicit) {
-            quote! {
-                decoder.decode_sequence(#tag, |decoder| {
-                    #encode
-                })
-            }
-        } else {
-            encode
         }
     }
 
@@ -312,6 +317,7 @@ impl<'a> VariantConfig<'a> {
                 .fields
                 .iter()
                 .enumerate()
+                .filter(|(_, f)| FieldConfig::new(f, self.container_config).is_option_type())
                 .map(|(i, f)| FieldConfig::new(f, self.container_config).tag_tree(i));
 
             match self.variant.fields {
@@ -319,10 +325,16 @@ impl<'a> VariantConfig<'a> {
                     quote!(#crate_root::TagTree::Leaf(<() as #crate_root::AsnType>::TAG),)
                 }
                 syn::Fields::Named(_) => {
+                    let error_message = format!("{}'s fields is not a valid \
+                        order of ASN.1 tags, ensure that your field's tags and \
+                        OPTIONALs are correct.",
+                            self.variant.ident
+                        );
+
                     quote!({
                         const FIELD_LIST: &'static [#crate_root::TagTree] = &[#(#field_tags,)*];
                         const FIELD_TAG_TREE: #crate_root::TagTree = #crate_root::TagTree::Choice(FIELD_LIST);
-                        #crate_root::sa::const_assert!(FIELD_TAG_TREE.is_unique());
+                        const _: () = assert!(FIELD_TAG_TREE.is_unique(), #error_message);
                         #crate_root::TagTree::Leaf(#crate_root::Tag::SEQUENCE)
                     },)
                 }
@@ -368,13 +380,10 @@ impl<'a> FieldConfig<'a> {
                     tag = Tag::from_meta(item);
                 } else if path.is_ident("default") {
                     default = Some(match item {
-                        syn::Meta::List(list) => list
-                            .nested
-                            .iter()
-                            .cloned()
-                            .filter_map(unested_meta)
-                            .map(|m| m.path().clone())
-                            .next(),
+                        syn::Meta::NameValue(value) => match &value.lit {
+                            syn::Lit::Str(lit_str) => lit_str.parse().map(Some).unwrap(),
+                            _ => panic!("Unsupported type for default."),
+                        },
                         _ => None,
                     });
                 }
@@ -411,9 +420,10 @@ impl<'a> FieldConfig<'a> {
         };
 
         if self.default.is_some() {
+            let ty = &self.field.ty;
             let default_fn = match self.default.as_ref().unwrap() {
-                Some(path) => quote!(#path),
-                None => quote!(<_>::default()),
+                Some(path) => quote!(#path ()),
+                None => quote!(<#ty>::default()),
             };
 
             quote! {
@@ -429,7 +439,7 @@ impl<'a> FieldConfig<'a> {
     pub fn decode(&self, name: &syn::Ident, context: usize) -> proc_macro2::TokenStream {
         let crate_root = &self.container_config.crate_root;
         let or_else = match self.default {
-            Some(Some(ref path)) => quote! { .unwrap_or_else(#path) },
+            Some(Some(ref path)) => quote! { .unwrap_or_else(|_| #path ()) },
             Some(None) => quote! { .unwrap_or_default() },
             None => {
                 let ident = format!(
@@ -492,10 +502,3 @@ impl<'a> FieldConfig<'a> {
             .is_option_type(&self.field.ty)
     }
 }
-
-fn unested_meta(nm: syn::NestedMeta) -> Option<syn::Meta> {
-    match nm {
-        syn::NestedMeta::Meta(m) => Some(m),
-        _ => None,
-    }
-}
